@@ -13,16 +13,14 @@
 
 #ifdef SunOS
 #define DEFAULT_MAIL_COMMAND "/usr/bin/mailx"
-#define COMPRESS_COMMAND "/usr/local/bin/gzip"
-#define UNCOMPRESS_COMMAND "/usr/local/bin/gunzip"
 #define STATEFILE "/var/log/logrotate.status"
 #endif
 
 #ifdef __NetBSD__
-   #define DEFAULT_MAIL_COMMAND "/usr/bin/mail -s"
-   #define COMPRESS_COMMAND "/usr/bin/gzip"
-   #define UNCOMPRESS_COMMAND "/usr/bin/gunzip"
-   #define STATEFILE "/var/log/logrotate.status"
+#define DEFAULT_MAIL_COMMAND "/usr/bin/mail -s"
+#define COMPRESS_COMMAND "/usr/bin/gzip"
+#define UNCOMPRESS_COMMAND "/usr/bin/gunzip"
+#define STATEFILE "/var/log/logrotate.status"
 #endif
 
 /*
