@@ -1,4 +1,4 @@
-#include <sys/queue.h>
+#include "queue.h"
 /* alloca() is defined in stdlib.h in NetBSD */
 #ifndef __NetBSD__
 #include <alloca.h>
@@ -72,6 +72,8 @@ extern int asprintf(char **str, const char *fmt, ...);
 #define DOEXIT exit
 #endif
 
+#define ERROR_STOP_ROTATION 1
+#define ERROR_CONTINUE_ROTATION 2
 
 struct logState {
     char *fn;
@@ -117,20 +119,24 @@ static int globerr(const char *pathname, int theerr)
     return 1;
 }
 
-#if defined(HAVE_STRPTIME) && defined(HAVE_QSORT_R)
-static int compGlobResult(const void *result1, const void *result2, void *data)  {
+#if defined(HAVE_STRPTIME) && defined(HAVE_QSORT)
+
+/* We could switch to qsort_r to get rid of this global variable,
+ * but qsort_r is not portable enough (Linux vs. *BSD vs ...)... */
+static struct compData _compData;
+
+static int compGlobResult(const void *result1, const void *result2)  {
 	struct tm time;
 	time_t t1, t2;
-	struct compData *d = (struct compData *) data;
 	const char *r1 = *(const char **)(result1);
 	const char *r2 = *(const char **)(result2);
 
 	memset(&time, 0, sizeof(struct tm));
-	strptime(r1 + d->prefix_len, d->dformat, &time);
+	strptime(r1 + _compData.prefix_len, _compData.dformat, &time);
 	t1 = mktime(&time);
 
 	memset(&time, 0, sizeof(struct tm));
-	strptime(r2 + d->prefix_len, d->dformat, &time);
+	strptime(r2 + _compData.prefix_len, _compData.dformat, &time);
 	t2 = mktime(&time);
 
 	if (t1 < t2) return -1;
@@ -139,10 +145,13 @@ static int compGlobResult(const void *result1, const void *result2, void *data)
 }
 
 static void sortGlobResult(glob_t *result, int prefix_len, const char *dformat) {
-	struct compData d;
-	d.prefix_len = prefix_len;
-	d.dformat = dformat;
-	qsort_r(result->gl_pathv, result->gl_pathc, sizeof(char *), compGlobResult, &d);
+	if (!dformat || *dformat == '\0') {
+		return;
+	}
+
+	_compData.prefix_len = prefix_len;
+	_compData.dformat = dformat;
+	qsort(result->gl_pathv, result->gl_pathc, sizeof(char *), compGlobResult);
 }
 #else
 static void sortGlobResult(glob_t *result, int prefix_len, const char *dformat) {
@@ -744,7 +753,7 @@ static int copyTruncate(char *currLog, char *saveLog, struct stat *sb,
 	if ((fdcurr = open(currLog, ((flags & LOG_FLAG_COPY) ? O_RDONLY : O_RDWR) | O_NOFOLLOW)) < 0) {
 	    message(MESS_ERROR, "error opening %s: %s\n", currLog,
 		    strerror(errno));
-	    return 1;
+	    return ERROR_CONTINUE_ROTATION;
 	}
 #ifdef WITH_SELINUX
 	if (selinux_enabled) {
@@ -1008,7 +1017,7 @@ int prerotateSingleLog(struct logInfo *log, int logNum, struct logState *state,
 #define DATEEXT_LEN 64
 #define PATTERN_LEN (DATEEXT_LEN * 2)
 	char dext_str[DATEEXT_LEN];
-	char dformat[DATEEXT_LEN];
+	char dformat[DATEEXT_LEN] = "";
 	char dext_pattern[PATTERN_LEN];
 	char *dext;
 
@@ -1086,12 +1095,12 @@ int prerotateSingleLog(struct logInfo *log, int logNum, struct logState *state,
 				switch (*(dext + 1)) {
 					case 'Y':
 						strncat(dext_pattern, "[0-9][0-9]",
-								sizeof(dext_pattern) - strlen(dext_pattern));
+								sizeof(dext_pattern) - strlen(dext_pattern) - 1);
 						j += 10; /* strlen("[0-9][0-9]") */
 					case 'm':
 					case 'd':
 						strncat(dext_pattern, "[0-9][0-9]",
-								sizeof(dext_pattern) - strlen(dext_pattern));
+								sizeof(dext_pattern) - strlen(dext_pattern) - 1);
 						j += 10;
 						if (j >= (sizeof(dext_pattern) - 1)) {
 							message(MESS_ERROR, "Date format %s is too long\n",
@@ -1106,7 +1115,7 @@ int prerotateSingleLog(struct logInfo *log, int logNum, struct logState *state,
 						/* End of year 2293 this pattern does not work. */
 						strncat(dext_pattern,
 								"[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9]",
-								sizeof(dext_pattern) - strlen(dext_pattern));
+								sizeof(dext_pattern) - strlen(dext_pattern) - 1);
 						j += 50;
 						if (j >= (sizeof(dext_pattern) - 1)) {
 							message(MESS_ERROR, "Date format %s is too long\n",
@@ -1423,6 +1432,7 @@ int rotateSingleLog(struct logInfo *log, int logNum, struct logState *state,
 #ifdef WITH_SELINUX
 	security_context_t savedContext = NULL;
 #endif
+	char *tmpFilename = NULL;
 
     if (!state->doRotate)
 	return 0;
@@ -1491,18 +1501,40 @@ int rotateSingleLog(struct logInfo *log, int logNum, struct logState *state,
 			if (!ACL_NOT_WELL_SUPPORTED(errno)) {
 				message(MESS_ERROR, "getting file ACL %s: %s\n",
 					log->files[logNum], strerror(errno));
-				hasErrors = 1;
+				hasErrors |= 1;
 			}
 		}
 #endif /* WITH_ACL */
-		message(MESS_DEBUG, "renaming %s to %s\n", log->files[logNum],
-		    rotNames->finalName);
-	    if (!debug && !hasErrors &&
-		rename(log->files[logNum], rotNames->finalName)) {
-		message(MESS_ERROR, "failed to rename %s to %s: %s\n",
-			log->files[logNum], rotNames->finalName,
-			strerror(errno));
-			hasErrors = 1;
+		if (log->flags & LOG_FLAG_TMPFILENAME) {
+			if (asprintf(&tmpFilename, "%s%s", log->files[logNum], ".tmp") < 0) {
+				message(MESS_FATAL, "could not allocate tmpFilename memory\n");
+				return 1;
+			}
+
+			message(MESS_DEBUG, "renaming %s to %s\n", log->files[logNum],
+				tmpFilename);
+			if (!debug && !hasErrors && rename(log->files[logNum], tmpFilename)) {
+			message(MESS_ERROR, "failed to rename %s to %s: %s\n",
+				log->files[logNum], tmpFilename,
+				strerror(errno));
+				if (errno == ENOENT) {
+					hasErrors |= ERROR_CONTINUE_ROTATION;
+				}
+				else {
+					hasErrors |= 1;
+				}
+			}
+		}
+		else {
+			message(MESS_DEBUG, "renaming %s to %s\n", log->files[logNum],
+				rotNames->finalName);
+			if (!debug && !hasErrors &&
+			rename(log->files[logNum], rotNames->finalName)) {
+				message(MESS_ERROR, "failed to rename %s to %s: %s\n",
+					log->files[logNum], tmpFilename,
+					strerror(errno));
+					hasErrors |= 1;
+			}
 	    }
 
 	    if (!log->rotateCount) {
@@ -1555,7 +1587,7 @@ int rotateSingleLog(struct logInfo *log, int logNum, struct logState *state,
 			}
 #endif
 			if (fd < 0)
-				hasErrors = 1;
+				hasErrors |= 1;
 			else {
 				close(fd);
 			}
@@ -1571,8 +1603,9 @@ int rotateSingleLog(struct logInfo *log, int logNum, struct logState *state,
 #endif
 
 	if (!hasErrors
-	    && log->flags & (LOG_FLAG_COPYTRUNCATE | LOG_FLAG_COPY)) {
-	    hasErrors =
+	    && log->flags & (LOG_FLAG_COPYTRUNCATE | LOG_FLAG_COPY)
+		&& !(log->flags & LOG_FLAG_TMPFILENAME)) {
+	    hasErrors |=
 		copyTruncate(log->files[logNum], rotNames->finalName,
 			     &state->sb, log->flags);
 	}
@@ -1596,7 +1629,22 @@ int postrotateSingleLog(struct logInfo *log, int logNum, struct logState *state,
     if (!state->doRotate)
 	return 0;
 
-    if ((log->flags & LOG_FLAG_COMPRESS) &&
+	if (!hasErrors && log->flags & LOG_FLAG_TMPFILENAME) {
+		char *tmpFilename = NULL;
+		if (asprintf(&tmpFilename, "%s%s", log->files[logNum], ".tmp") < 0) {
+			message(MESS_FATAL, "could not allocate tmpFilename memory\n");
+			return 1;
+		}
+	    hasErrors =
+		copyTruncate(tmpFilename, rotNames->finalName,
+			     &state->sb, LOG_FLAG_COPY);
+		message(MESS_DEBUG, "removing tmp log %s \n", tmpFilename);
+		if (!debug && !hasErrors) {
+			unlink(tmpFilename);
+		}
+	}
+
+    if (!hasErrors && (log->flags & LOG_FLAG_COMPRESS) &&
 	!(log->flags & LOG_FLAG_DELAYCOMPRESS)) {
 	hasErrors = compressLogFile(rotNames->finalName, log, &state->sb);
     }
@@ -1632,6 +1680,7 @@ int rotateLogSet(struct logInfo *log, int force)
     int i, j;
     int hasErrors = 0;
     int logHasErrors[log->numFiles];
+	int logHasScriptErrors[log->numFiles];
     int numRotated = 0;
     struct logState **state;
     struct logNames **rotNames;
@@ -1703,6 +1752,7 @@ int rotateLogSet(struct logInfo *log, int force)
     for (i = 0; i < log->numFiles; i++) {
 	logHasErrors[i] = findNeedRotating(log, i, force);
 	hasErrors |= logHasErrors[i];
+	logHasScriptErrors[i] = 0;
 
 	/* sure is a lot of findStating going on .. */
 	if ((findState(log->files[i]))->doRotate)
@@ -1771,7 +1821,8 @@ int rotateLogSet(struct logInfo *log, int force)
 				"error running non-shared prerotate script "
 				"for %s of '%s'\n", log->files[j], log->pattern);
 		    }
-		    logHasErrors[j] = 1;
+		    logHasScriptErrors[j] = 1;
+			logHasErrors[j] = 1;
 		    hasErrors = 1;
 		}
 	    }
@@ -1780,8 +1831,7 @@ int rotateLogSet(struct logInfo *log, int force)
 	for (i = j;
 	     ((log->flags & LOG_FLAG_SHAREDSCRIPTS) && i < log->numFiles)
 	     || (!(log->flags & LOG_FLAG_SHAREDSCRIPTS) && i == j); i++) {
-	    if (! ( (logHasErrors[i] && !(log->flags & LOG_FLAG_SHAREDSCRIPTS))
-		   || (hasErrors && (log->flags & LOG_FLAG_SHAREDSCRIPTS)) ) ) {
+	    if (!logHasScriptErrors[i] && !(logHasErrors[i] & ERROR_STOP_ROTATION)) {
 		logHasErrors[i] |=
 		    rotateSingleLog(log, i, state[i], rotNames[i]);
 		hasErrors |= logHasErrors[i];
@@ -1790,7 +1840,7 @@ int rotateLogSet(struct logInfo *log, int force)
 
 	if (log->post
 		&& (!(
-			((logHasErrors[j] || !state[j]->doRotate) && !(log->flags & LOG_FLAG_SHAREDSCRIPTS))
+			(((logHasErrors[j] & ERROR_STOP_ROTATION) || !state[j]->doRotate) && !(log->flags & LOG_FLAG_SHAREDSCRIPTS))
 			|| (hasErrors && (log->flags & LOG_FLAG_SHAREDSCRIPTS))
 		))
 	) {
@@ -1809,7 +1859,8 @@ int rotateLogSet(struct logInfo *log, int force)
 				"error running non-shared postrotate script "
 				"for %s of '%s'\n", log->files[j], log->pattern);
 		    }
-		    logHasErrors[j] = 1;
+		    logHasScriptErrors[j] = 1;
+			logHasErrors[j] = 1;
 		    hasErrors = 1;
 		}
 	    }
@@ -1818,8 +1869,7 @@ int rotateLogSet(struct logInfo *log, int force)
 	for (i = j;
 	     ((log->flags & LOG_FLAG_SHAREDSCRIPTS) && i < log->numFiles)
 	     || (!(log->flags & LOG_FLAG_SHAREDSCRIPTS) && i == j); i++) {
-	    if (! ( (logHasErrors[i] && !(log->flags & LOG_FLAG_SHAREDSCRIPTS))
-		   || (hasErrors && (log->flags & LOG_FLAG_SHAREDSCRIPTS)) ) ) {
+		if (!logHasScriptErrors[i] && !(logHasErrors[i] & ERROR_STOP_ROTATION)) {
 		logHasErrors[i] |=
 		    postrotateSingleLog(log, i, state[i], rotNames[i]);
 		hasErrors |= logHasErrors[i];
